﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using ArcToolkitCLI.Commands.Options;
using ArcToolkitCLI.Util;
using CommandLine;
using Squish;

namespace ArcToolkitCLI.Commands.Converters
{
    [Verb("tex", HelpText = "Convert TEX files into PNG and back")]
    internal class TexConverter : BaseConverter
    {
        private const string TEX_TAG = "CM3D2_TEX";
        private const int TEX_VERSION = 1010;

        private readonly byte[] pngHeader =
        {
            137,
            80,
            78,
            71,
            13,
            10,
            26,
            10
        };

        private readonly Dictionary<TextureFormat, Action<string, byte[], int, int, TextureFormat>> textureLoaders;

        public TexConverter()
        {
            Converters = new Dictionary<string, Func<string, int>>
            {
                [".tex"] = TexToPng,
                [".png"] = PngToTex
            };
            
            textureLoaders = new Dictionary<TextureFormat, Action<string, byte[], int, int, TextureFormat>>
            {
                [TextureFormat.Argb32] = ConvertFromPng,
                [TextureFormat.Rgb24] = ConvertFromPng,
                [TextureFormat.Dxt1] = ConvertFromDxt,
                [TextureFormat.Dxt5] = ConvertFromDxt
            };
        }

        private static bool ByteEquals(IReadOnlyList<byte> b1, IReadOnlyList<byte> b2)
        {
            var len = Math.Min(b1.Count, b2.Count);

            for (var i = 0; i < len; i++)
                if (b1[i] != b2[i])
                    return false;
            return true;
        }

        private int PngToTex(string file)
        {
            var img = File.ReadAllBytes(file);

            if (!ByteEquals(img, pngHeader))
            {
                Console.WriteLine($"File {file} is not a PNG file!");
                return 1;
            }

            var w = BitConverter.ToInt32(img, pngHeader.Length + 4 + 4);
            var h = BitConverter.ToInt32(img, pngHeader.Length + 4 + 4 + 4);
            using var bw =
                new BinaryWriter(File.Create(Path.Combine(Output, $"{Path.GetFileNameWithoutExtension(file)}.tex")));
            bw.Write(TEX_TAG);
            bw.Write(TEX_VERSION);
            bw.Write(string.Empty);
            bw.Write(w);
            bw.Write(h);
            bw.Write((int)TextureFormat.Argb32);
            bw.Write(img.Length);
            bw.Write(img);
            return 0;
        }

        private int TexToPng(string file)
        {
            using var br = new BinaryReader(File.OpenRead(file));
            var tag = br.ReadString();

            if (tag != TEX_TAG)
            {
                Console.WriteLine($"File {file} is not a valid TEX file!");
                return 1;
            }

            var version = br.ReadInt32();
            br.ReadString();
            var width = 0;
            var height = 0;

            var format = TextureFormat.Argb32;

            if (version >= 1010)
            {
                width = br.ReadInt32();
                height = br.ReadInt32();
                format = (TextureFormat)br.ReadInt32();
            }

            if (!Enum.IsDefined(typeof(TextureFormat), format))
            {
                Console.WriteLine($"File {file} has unsupported texture format: {format}");
                return 1;
            }

            var size = br.ReadInt32();
            var data = new byte[size];
            br.Read(data, 0, size);

            if (version == 1000)
            {
                width = BitConverter.ToInt32(data, pngHeader.Length + 4 + 4);
                height = BitConverter.ToInt32(data, pngHeader.Length + 4 + 4 + 4);
            }

            if (textureLoaders.TryGetValue(format, out var saveTex))
                saveTex(file, data, width, height, format);
            else
            {
                Console.WriteLine($"File {file} uses format {format} that is not supported!");
                return 1;
            }

            return 0;
        }

        private void ConvertFromPng(string file, byte[] data, int width, int height, TextureFormat format)
        {
            var ms = new MemoryStream(data);
            var img = Image.FromStream(ms);
            img.Save(Path.Combine(Output, $"{Path.GetFileNameWithoutExtension(file)}.png"));
            img.Dispose();
            ms.Dispose();
        }

        private void ConvertFromDxt(string file, byte[] data, int width, int height, TextureFormat format)
        {
            var squishFlags =
                format switch 
                { 
                    TextureFormat.Dxt1 => SquishFlags.kDxt1, 
                    TextureFormat.Dxt5 => SquishFlags.kDxt5,
                    _ => throw new ArgumentException($"Invalid DXT texture format: {format}", nameof(format)) 
                };

            var outData = new byte[width * height * 4];
            Squish.Squish.DecompressImage(outData, width, height, ref data, squishFlags);

            for (var i = 0; i < width * height; i++)
            {
                var r = outData[i * 4];
                outData[i * 4] = outData[i * 4 + 2];
                outData[i * 4 + 2] = r;
            }

            var gch = GCHandle.Alloc(outData, GCHandleType.Pinned);
            var img = new Bitmap(width, height, width * 4, PixelFormat.Format32bppArgb, gch.AddrOfPinnedObject());
            img.RotateFlip(RotateFlipType.RotateNoneFlipY);

            img.Save(Path.Combine(Output, $"{Path.GetFileNameWithoutExtension(file)}.png"));

            img.Dispose();
            gch.Free();
        }

        private enum TextureFormat
        {
            Argb32 = 5,
            Rgb24 = 3,
            Dxt1 = 10,
            Dxt5 = 12
        }
    }
}